/**
    Copyright (C) 2002-2003  Together

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

package org.relique.jdbc.csv;

import java.io.*;
import java.util.*;
import java.sql.SQLException;

/**
 * This class is a helper class that handles the reading and parsing of data
 * from a .csv file.
 *
 * @author     Sinisa Milosevic
 * @author     Zoran Milakovic
 */

public class CsvWriter
{
  public RandomAccessFile randomCsvFile;
  private File file;
  private String[] columnNames;
  private String[] columnTypes;
  private String[] columns;
  private java.lang.String buf = null;
  private char separator = CsvDriver.DEFAULT_SEPARATOR;
  private long maxFileSize = CsvDriver.DEFAULT_FILE_MAXSIZE;
  private String extension = CsvDriver.DEFAULT_EXTENSION;
  private String tableName;
  private String fileName;
  private int counter;
  private long current;
  private long endLine;


  /**
   * Used with statement.
   *
   * @param fileName
   * @param separator
   * @param extension
   * @param maxFileSize
   * @throws java.lang.Exception
   */
   public CsvWriter(
      String fileName,
      char separator,
      String extension,
      long maxFileSize
      )
      throws java.lang.Exception
  {
    this.separator = separator;
    this.extension = extension;
    this.maxFileSize = maxFileSize;
    if( fileName != null ) {
      this.fileName = fileName;
      this.randomCsvFile = new RandomAccessFile(fileName, "rw");
      this.file = new File(fileName);
      String headerLine = this.randomCsvFile.readLine();
      columnNames = parseCsvLineAsHeader(headerLine);
    }
    counter = 1;
    current = 0;
    endLine=0;
  }


  /**
   * When use split files, this is used when file name is changed.
   * @param name
   * @throws Exception
   */
  public void setFileName(String name) throws Exception {
    this.fileName=name;
    if( this.randomCsvFile != null )
      this.randomCsvFile.close();
    this.file = new File( this.fileName );
    this.randomCsvFile = new RandomAccessFile(this.fileName,"rw");
  }


  public void fillTableColumnNames() throws java.lang.Exception {
    String headerLine = this.randomCsvFile.readLine();
    this.columnNames = parseCsvLineAsHeader(headerLine);
  }

  /**
   * Gets the columnNames attribute of the CsvReader object
   *
   * @return The columnNames value
   * @since
   */
  public String[] getColumnNames()
  {
    return columnNames;
  }

  private String getNextFileName(String currentFileName) {
    String newName = "";
    String number = "";
//name without extension
    String currentFileExtension = currentFileName.substring(currentFileName.lastIndexOf("."), currentFileName.length());
    currentFileName = currentFileName.substring(0, currentFileName.lastIndexOf("."));
    if( currentFileExtension.endsWith(CsvDriver.FILE_NAME_EXT) ) {
       number += currentFileName.substring(currentFileName.length()-3, currentFileName.length());
       long num = Long.valueOf(number).longValue()+1;
       if( num >= 100 && num < 1000 )
         number = String.valueOf( num );
       else if ( num >= 10 && num < 100 )
         number = "0"+String.valueOf( num );
       else if ( num > 1 && num < 10 )
         number = "00"+String.valueOf( num );
       currentFileName = currentFileName.substring(0, currentFileName.length()-3);
       newName = currentFileName + number + currentFileExtension;
    } else {
      newName = currentFileName.toUpperCase() + "001" + this.extension + CsvDriver.FILE_NAME_EXT;
    }
    return newName;
  }


  public String getTableName() {
    if(tableName != null)
      return tableName;

    int lastSlash = 0;
    for(int i = fileName.length()-1; i >= 0; i--)
      if(fileName.charAt(i) == '/' || fileName.charAt(i) == '\\') {
    lastSlash = i;
    break;
      }
      tableName = fileName.substring(lastSlash+1, fileName.length() - 4);
      return tableName;
  }

  /**
   * Get the value of the column at the specified index.
   *
   * @param  columnIndex  Description of Parameter
   * @return              The column value
   * @since
   */

  public String getColumn(int columnIndex)
  {
    return columns[columnIndex];
  }

  /**
   * Get value from column at specified name.
   * If the column name is not found, throw an error.
   *
   * @param  columnName     Description of Parameter
   * @return                The column value
   * @exception  SQLException  Description of Exception
   * @since
   */

  public String getColumn(String columnName) throws SQLException
  {
    for (int loop = 0; loop < columnNames.length; loop++)
    {
      if (columnName.equalsIgnoreCase(columnNames[loop]) || columnName.equalsIgnoreCase(getTableName() + "." + columnNames[loop]))
      {
      return getColumn(loop);
    }
    }
    throw new SQLException("Column '" + columnName + "' not found.");
  }


  /**
   *Description of the Method
   *
   * @return                Description of the Returned Value
   * @exception  SQLException  Description of Exception
   * @since
   */
  private long lastCurrent = -1;
  public boolean next() throws SQLException, IOException {
    columns = new String[columnNames.length];
    String dataLine = null;
    if( lastCurrent == -1 ) {
      current = randomCsvFile.getFilePointer();
    } else {
      current = lastCurrent;
    }

    try {
      if (buf != null) {
        // The buffer is not empty yet, so use this first.
        dataLine = buf;
        buf = null;
      } else {
        // read new line of data from input.
        if( lastCurrent != -1 )
          randomCsvFile.seek(lastCurrent);
        dataLine = this.randomCsvFile.readLine();
      }
      if (dataLine == null) {
        String nextFileName = getNextFileName(this.fileName);
          if( new File(nextFileName).exists() ) {
            this.fileName = nextFileName;
            if( randomCsvFile != null )
            	randomCsvFile.close();
            randomCsvFile = new RandomAccessFile(this.fileName, "rw");
            counter = 1;
            current = 0;
            endLine=0;
            //skip header
//            dataLine = input.readLine();
            dataLine = randomCsvFile.readLine();
          } else {
            randomCsvFile.close();
            return false;
          }
      }

    } catch (IOException e) {
      throw new SQLException(e.toString());
    }
    columns = parseCsvLine(dataLine);
    endLine = randomCsvFile.getFilePointer();
    return true;
  }


  /**
   *Description of the Method
   *
   * @since
   */
  public void close()
  {
    try
    {
      this.file = null;
      this.randomCsvFile.close();
      buf = null;
    }
    catch (Exception e)
    {
    }
  }


  /**
  *
  * @param line
  * @return array with values or column names.
  * @throws SQLException
  */
 protected String[] parseCsvLine(String line) throws SQLException
 {
   ArrayList values = new ArrayList();
   boolean inQuotedString = false;
   String value = "";
   String orgLine = line;
   int currentPos = 0;
   int fullLine = 0;
   int currentColumn = 0;
   char currentChar;

     line += separator;
     long lineLength = line.length();
     while (fullLine == 0) {
       currentPos = 0;
       while (currentPos < lineLength) {

         //handle BINARY columns
        if( !(this.columnTypes.length <= currentColumn ) ) {
         if (this.columnTypes[currentColumn].equals(CsvDriver.BINARY_TYPE)) {
           String binaryValue = "";
           currentChar = line.charAt(currentPos);
           if (currentChar == ',') {
             values.add(binaryValue); //binary value is null;
             currentPos ++;
           }
           else if (currentChar == '"') {
             if (line.charAt(currentPos + 1) == '"') {
               values.add(binaryValue); //binary value is null
               currentPos = currentPos + 3;
             }
             else {
               // take all until next separator, and that is value
               // do not insert BinaryObject+index into line, just set right currentPos
               // and insert value into vector
               // binary value is always beteween quotes (")
               binaryValue = line.substring(currentPos);
               binaryValue = binaryValue.substring(1,
                                                   binaryValue.indexOf(separator) -
                                                   1);
               values.add(binaryValue);
               currentPos += binaryValue.length() + 3;
             }
           }
           //set currentColumn++
           currentColumn++;
           continue;
         }
        } else {
          throw new SQLException("Invalid csv format : file = "+new File(fileName).getAbsolutePath()+", line = "+line);
       }



//parse one by one character
         currentChar = line.charAt(currentPos);
         if (value.length() == 0 && currentChar == '"' && !inQuotedString) {
//enter here if we are at start of column value
           currentPos++;
           inQuotedString = true;
           continue;
         }

         if (currentChar == '"') {
//get next character
           char nextChar = line.charAt(currentPos + 1);
//if we have "", consider it as ", and add it to value
           if (nextChar == '"') {
             value += currentChar;
             currentPos++;
           }
           else {
//enter here if we are at end of column value
             if (!inQuotedString) {
               throw new SQLException("Unexpected '\"' in position " +
                                      currentPos + ". Line=" + orgLine);
             }
             if (inQuotedString && nextChar != separator) {
               throw new SQLException("Expecting " + separator +
                                      " in position " + (currentPos + 1) +
                                      ". Line=" + orgLine);
             }

//set currentPos to comma after value
             currentPos++;
//if value is empty string between double quotes consider it as empty string
//else if value is empty string between commas consider it as null value
             values.add(value);
             currentColumn++;
             value = "";
             inQuotedString = false;
           }
         }

         else {
//when we are at end of column value, and value is not inside of double quotes
           if (currentChar == separator) {
//when have separator in data
             if (inQuotedString) {
               value += currentChar;
             }
             else {
//if value is empty string between double quotes consider it as empty string
//else if value is empty string between commas consider it as null value
                if( value.equals("") )
                  value = null;
               values.add(value);
               currentColumn++;
               value = "";
             }
           }
           else {
             value += currentChar;
           }
         }

         currentPos++;
       } //end while

       if (inQuotedString) {
         // Remove extra , added at start
         value = value.substring(0, value.length() - 1);
         try {
           line = randomCsvFile.readLine();
         }
         catch (IOException e) {
           throw new SQLException(e.toString());
         }
       }
       else {
         fullLine = 1;
       }

     }// end while( fullLine == 0 )
   String[] retVal = new String[values.size()];
   values.toArray(retVal);

   return retVal;
 }


 /**
  *
  * @param line
  * @return array with values or column names.
  * @throws SQLException
  */
 protected String[] parseCsvLineAsHeader(String line) throws SQLException
 {
   Vector values = new Vector();
   ArrayList columnTypesList = new ArrayList();
   boolean inQuotedString = false;
   String value = "";
   String orgLine = line;
   int currentPos = 0;
   int fullLine = 0;

   while (fullLine == 0) {
     currentPos = 0;
     line += separator;
     while (currentPos < line.length()) {
       char currentChar = line.charAt(currentPos);
       if (value.length() == 0 && currentChar == '"' && !inQuotedString) {
         currentPos++;
         inQuotedString = true;
         continue;
       }
       if (currentChar == '"') {
         char nextChar = line.charAt(currentPos + 1);
         if (nextChar == '"') {
           value += currentChar;
           currentPos++;
         }
         else {
           if (!inQuotedString) {
             throw new SQLException("Unexpected '\"' in position " +
                                    currentPos + ". Line=" + orgLine);
           }
           if (inQuotedString && nextChar != separator) {
             throw new SQLException("Expecting " + separator + " in position " +
                                    (currentPos + 1) + ". Line=" + orgLine);
           }
           if (value.endsWith("-"+CsvDriver.BINARY_TYPE)) {
             columnTypesList.add(CsvDriver.BINARY_TYPE);
             value = value.substring(0,value.indexOf("-"+CsvDriver.BINARY_TYPE));
           }
           else
             columnTypesList.add(CsvDriver.VARCHAR_TYPE);
           values.add(value);
           value = "";
           inQuotedString = false;
           currentPos++;
         }
       }
       else {
         if (currentChar == separator) {
           if (inQuotedString) {
             value += currentChar;
           }
           else {
             if (value.endsWith("-"+CsvDriver.BINARY_TYPE)) {
               columnTypesList.add(CsvDriver.BINARY_TYPE);
               value = value.substring(0,
                                       value.indexOf("-"+CsvDriver.BINARY_TYPE));
             }
             else
               columnTypesList.add(CsvDriver.VARCHAR_TYPE);
             values.add(value);
             value = "";
           }
         }
         else {
           value += currentChar;
         }
       }
       currentPos++;
     }
     if (inQuotedString) {
       // Remove extra , added at start
       value = value.substring(0, value.length() - 1);
       try {
         line = randomCsvFile.readLine();
       }
       catch (IOException e) {
         throw new SQLException(e.toString());
       }
     }
     else {
       fullLine = 1;
     }
   }
   String[] retVal = new String[values.size()];
   values.copyInto(retVal);

   this.columnTypes = new String[columnTypesList.size()];
   columnTypesList.toArray(columnTypes);

   return retVal;

 }




  protected boolean newLine(String[] colNames, String[] colValues) throws IOException {
    String newLine="";
    boolean more = true;
    boolean createNewOutput = false;
//find out if file size is out of range, and if so create new file
    while(more) {
      if( (this.maxFileSize != -1) && (this.file.length() > this.maxFileSize) ) {
        String newTableName = this.getNextFileName(this.fileName);
        this.fileName = newTableName;
        this.createExtTable(this.columnNames, this.columnTypes, newTableName);
      } else {
        createNewOutput = true;
        more = false;
      }
    }
    try {
      if (createNewOutput) {
        this.randomCsvFile.close();
        this.randomCsvFile = new RandomAccessFile(this.fileName, "rw");
      }
    }catch(Exception e) {
      e.printStackTrace();
    }

    for(int i=0;i<columnNames.length; i++) {
      boolean find = false;
      for(int j=0;j<colNames.length; j++) {
        if(colNames[j].equalsIgnoreCase(columnNames[i])) {
//          if(colValues[j]==null || colValues[j].equals("null"))
          if(colValues[j]==null)
            newLine=newLine+separator;
          else	{
            newLine=newLine+"\""+colValues[j]+"\""+separator;
          }
          find = true;
        }
      }
      if(!find)
        newLine=newLine+separator;
    }
    if(!newLine.equals(""))
      newLine=newLine.substring(0,newLine.length()-1);

    long l = this.randomCsvFile.length();
    this.randomCsvFile.seek(l);
    this.randomCsvFile.writeBytes("\n"+newLine);
    this.randomCsvFile.close();

    return true;
  }

  protected boolean createTable(String[] colNames, String table) throws IOException {
    String newLine="";
    for(int i=0;i<colNames.length; i++) {
      newLine=newLine+"\""+colNames[i]+"\""+separator;
    }
    if(!newLine.equals(""))
      newLine=newLine.substring(0,newLine.length()-1);

    this.fileName = table;

    this.file = new File( this.fileName );

//    if( !this.file.exists() ) {
      if( this.randomCsvFile != null )
      	this.randomCsvFile.close();
      this.randomCsvFile = new RandomAccessFile(fileName,"rw");
      this.randomCsvFile.writeBytes(newLine);
      this.randomCsvFile.close();
//    }

    return true;
  }

  protected boolean createExtTable(String[] colNames, String[] colTypes, String table) throws IOException {
    String newLine="";
    for(int i=0;i<colNames.length; i++) {
      if(colTypes[i].equals(CsvDriver.BINARY_TYPE))
        newLine=newLine+"\""+colNames[i]+"-"+colTypes[i]+"\""+separator;
      else
        newLine=newLine+"\""+colNames[i]+"\""+separator;
    }
    if(!newLine.equals(""))
      newLine=newLine.substring(0,newLine.length()-1);

    this.fileName = table;

    this.file = new File( this.fileName );

    if( !this.file.exists() ) {
      RandomAccessFile temp = new RandomAccessFile(this.fileName,"rw");
      temp.writeBytes(newLine);
      temp.close();
    }

    return true;
  }


  protected boolean updateFields(
      String[] colNames,
      String[] colValues,
      String[] colWhereNames,
      String[] colWhereValues
      ) throws IOException, SQLException {

    boolean isUpdated=false;

    while( next() ) {
      isUpdated=false;
      counter++;
      boolean find = false;
      out:
      for(int i=0; i<colWhereNames.length; i++) {
//compare values
        if( ! Utils.compareValues( getColumn(colWhereNames[i]), colWhereValues[i] ) )
          break out;
        if(i==(colWhereNames.length-1)) {
          find = true;
        }
      }
//if there is no where clause
      if( colWhereNames.length == 0 )
        find = true;

//go to next line
      lastCurrent = randomCsvFile.getFilePointer();
      if(find) {
        for(int i=0; i < columnNames.length; i++) {
          for(int j=0; j<colNames.length; j++) {
            if(colNames[j].equalsIgnoreCase(columnNames[i]))
              columns[i]=colValues[j];
          }
        }
        String updatedLine = "";
        for(int i=0; i<columns.length; i++) {
          if(columns[i]==null)
            updatedLine=updatedLine+separator;
          else {
            updatedLine=updatedLine+"\""+columns[i]+"\""+separator;
          }

        }
        if(!updatedLine.equals("")) {
          randomCsvFile.seek(endLine);
          String line="";
          String newLine="";

          while(( newLine=randomCsvFile.readLine())!=null ){
            line+=newLine+"\n";
          }
          if(line!=null) {
            if(!line.equals(""))
              line=line.substring(0,line.length()-1);
          }
          updatedLine=updatedLine.substring(0,updatedLine.length()-1);
          randomCsvFile.seek(current);
          randomCsvFile.writeBytes(updatedLine);
          //go to next line
          lastCurrent = randomCsvFile.getFilePointer();
          if( randomCsvFile.getFilePointer() != randomCsvFile.length() ) {
            randomCsvFile.writeBytes("\n");
            //go to next line
            lastCurrent = randomCsvFile.getFilePointer();
            if(line!=null){
              if(!line.equals(""))
                randomCsvFile.writeBytes(line);
              randomCsvFile.setLength(randomCsvFile.getFilePointer());
            }
          }
          isUpdated = false;
        }
      }
    }
    return isUpdated;
  }

}
