/*
    Copyright (C) 2003  Together

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package org.webdocwf.util.xml;

//xml imports
import org.w3c.dom.Document;
import org.w3c.dom.NodeList;
import org.w3c.dom.Node;
import org.w3c.dom.Element;
import org.enhydra.xml.SearchElement;
import org.enhydra.xml.XMLDocumentFactory;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import java.sql.*;
import java.io.File;
import java.util.Properties;
import java.io.RandomAccessFile;
import java.util.ArrayList;

/**
 * Load existing XML file , creating DOM from file or creating
 * new DOM.Class has methods for insert,update,delete data from XML file and save new DOM in XML file.
 *
 * @author   Zoran Milakovic
*/

public class XmlWriter {

  /**
   * Document made from XML file, and in which will
   * be made changes.Document will be saved in XML file.
   */
  public static SearchElement searchDocumentStatic;
  private SearchElement searchDocument;
  private Document document;
  private static boolean isError = false;
  /**
   * If parameter is true document will be saved in xml file after each query,other wise
   * document will be saved when Connection.commit() method is called.
   */
  private boolean autoCommit = false;
  /**
   * Full path of the XML file.
   */
  private String fileName;

  /**
   * Constructor used when autoCommit is set to false or true.
   * @param fileName name of xml file.
   * @param isAutoCommit define is mod auto commit or not.
   */
  public XmlWriter(String fileName, boolean isAutoCommit) throws SQLException {
    try {
      XmlWriter.isError = false;
      this.autoCommit = isAutoCommit;
      DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
      DocumentBuilder builder = factory.newDocumentBuilder();
//check if database exist, and if not, create one
      this.fileName = fileName;
      File file = new File( fileName );
      if( !file.exists() ) {
        file.getParentFile().mkdirs();
        this.document = builder.newDocument();
        Element newDatabaseElement = document.createElement("database");
        newDatabaseElement.appendChild( document.createElement("dml") );
        document.insertBefore( newDatabaseElement , null );
        this.searchDocument = (SearchElement)SearchElement.newInstance( document );
        this.createDatabase();
      }

      if( isAutoCommit ) {
        this.document = builder.parse( file );
        this.searchDocument = (SearchElement)SearchElement.newInstance( document );
      } else {
        if( this.searchDocumentStatic == null ) {
          this.document = builder.parse( file );
          this.searchDocument = (SearchElement)SearchElement.newInstance( this.document );
        } else {
          this.searchDocument = searchDocumentStatic;
        }
      }

      /* set document to new value, to prevent exist of two same documents
         in database to decrease memory usage */
      this.document = builder.newDocument();

      } catch( Exception e ) { throw new SQLException("Error in constructor : "+e.getMessage()); }
  }

  /**
   * Adds sql statement CREATE TABLE in XML file.
   * Method will throw SQLException with appropriate message if table already exist
   *
   * @param sqlStatement CREATE TABLE statement which will be add into XML file
   * @param tableName name of table which will be created
   * @throws SQLException
   */

  protected void createTable(String sqlStatement, String tableName) throws SQLException {
    try {
      boolean allreadyExist = false;
      NodeList sqlStatements = searchDocument.getSubElementsByTagName("ddl");
      XmlSqlParser parser = new XmlSqlParser();
      for( int i = 0; i < sqlStatements.getLength(); i++ ) {
        Node node = sqlStatements.item(i);
        parser.parse( node.getFirstChild().toString() );
        if ( parser.getTableName().equalsIgnoreCase( tableName ) ) {
          allreadyExist = true;
        }
      }
      NodeList existingTables = searchDocument.getSubElementsByTagName("dml/"+tableName);
      if( existingTables.getLength() != 0 ) {
        allreadyExist = true;
      }
      if( allreadyExist ) {
        this.isError = true;
        throw new SQLException("Table with specified name already exist ! ");
      } else {
        Element newDDLElement = document.createElement("ddl");
        newDDLElement.appendChild( document.createTextNode( sqlStatement ) );
        SearchElement newDDL = new SearchElement( newDDLElement );
        NodeList dml = searchDocument.getSubElementsByTagName("dml");
        Element before = null;
        if( dml.getLength() != 0 )
          before = (Element)dml.item(0);
        searchDocument.insertBefore( newDDL , before );
      }
      saveDOM();
    } catch( Exception e ) {
      this.isError = true;
      throw new SQLException("Error in creating table : "+e.getMessage());
    }
  }

  /**
   * Delete row(s) from XML file.
   *
   * @param tableName name of table from which will be deleted rows.
   * @param whereColumnNames names of columns in WHERE clause.
   * @param whereColumnValues values of columns in WHERE clause.
   * @throws SQLException
   */
  protected void delete(String tableName, String[] whereColumnNames , String[] whereColumnValues) throws SQLException {
    try {
      NodeList tableRows = searchDocument.getSubElementsByTagName("dml/"+tableName);
      //check if table row match conditions
      for(int i = 0; i < tableRows.getLength(); i++) {
        boolean isMatch = true;
        if( whereColumnNames != null && whereColumnValues != null ) {
          for(int k = 0; k < whereColumnNames.length; k++) {
            NodeList columns = ( (SearchElement)tableRows.item(i) ).getSubElementsByCondition(whereColumnNames[k]+"="+whereColumnValues[k]);
            if( columns.getLength() == 0 )
              isMatch = false;
          }
        }
        if( whereColumnNames.length == 0 )
          isMatch=true;
        if( isMatch ) {
          //deleting row from XML database
          SearchElement parent = new SearchElement( tableRows.item(i).getParentNode() );
          parent.removeChild( tableRows.item(i) );
        }
      }
      saveDOM();
      } catch( Exception e ) {
        this.isError = true;
        throw new SQLException("Error in delete data: "+e.getMessage());
      }
  }

  /**
   * Delete table from XML file.
   *
   * @param tableName name of table which will be deleted.
   * @throws SQLException
   */
  protected void dropTable( String tableName ) throws SQLException {
    try {
      //delete data
      NodeList tableRows = searchDocument.getSubElementsByTagName("dml/"+tableName);
      for( int i = 0; i < tableRows.getLength(); i++ ) {
        SearchElement parent = new SearchElement( tableRows.item(i).getParentNode() );
        parent.removeChild( tableRows.item(i) );
      }
      //delete CREATE TABLE statement if exist
      NodeList sqlStatements = searchDocument.getSubElementsByTagName("ddl");
      XmlSqlParser parser = new XmlSqlParser();
      for( int i = 0; i < sqlStatements.getLength(); i++ ) {
        Node node = sqlStatements.item(i);
        try {
          parser.parse( node.getFirstChild().toString() );
        } catch(Exception e) {
          this.isError = true;
          throw new SQLException("Error in parsing statement : "+e.getMessage());
        }
        if ( parser.getTableName().equalsIgnoreCase( tableName ) && parser.sqlType.equalsIgnoreCase( parser.CREATE_TABLE ) ) {
          SearchElement parent = new SearchElement( sqlStatements.item(i).getParentNode() );
          parent.removeChild( sqlStatements.item(i) );
        }
      }
      saveDOM();
      } catch( Exception e ) {
        this.isError = true;
        throw new SQLException("Error in drop table : "+e.getMessage());
      }
  }


  /**
   * Insert row in XML file.
   *
   * @param tableName name of table in which will be added rows.
   * @param columnNames names of columns in which will be added data.
   * @param columnValues value which will be insert into table.
   * @throws SQLException
   */
  protected void insert(String tableName, String[] columnNames , String[] columnValues) throws SQLException {
    try {
      String[] allColumnNames = (String[])this.getTableProperties( tableName ).get(0);
      boolean hasCreateTable =
          !this.getTableProperties( tableName ).get(1).toString().equalsIgnoreCase("NO CREATE TABLE");
      if( hasCreateTable ) {
        String[] primaryKeys = (String[])this.getTableProperties( tableName ).get(1);
        //check if column is primarykey and if is it duplicate value
        for(int k = 0; k < columnNames.length; k++) {
          boolean isPrimarykey = false;
          for(int i = 0; i < primaryKeys.length; i++) {
            if( columnNames[k].equals( primaryKeys[i] ) )
              isPrimarykey = true;
          }
          if( isPrimarykey ) {
            NodeList columns = searchDocument.getSubElementsByCondition("dml/"+tableName+"/"+columnNames[k]+"="+columnValues[k]);
            if( columns.getLength() != 0 ) {
              this.isError = true;
              throw new SQLException("Can not insert duplicate value in primary key column "+columnNames[k]+" !");
            }
          }
        }
      }
      //create new table
      Element newTable = document.createElement(tableName);
      if( hasCreateTable ) {
        //has CREATE TABLE statement
        Element newColumn;
        for( int i = 0; i < columnNames.length; i++ ) {
          newColumn = document.createElement( columnNames[i] );
          if( columnValues[i].equals("null") ) {
            String[] notNullCols = (String[])this.getTableProperties( tableName ).get(2);
            for( int ii=0; ii < notNullCols.length; ii++ ) {
              if( notNullCols[ii].equalsIgnoreCase(columnNames[i]) )
                throw new SQLException("Column '" + columnNames[i] + "' can not be NULL.");
            }
            //do not add empty column, because there are CREATE TABLE statement
            continue;
          }
          else {
            newColumn.appendChild( document.createTextNode( columnValues[i] ) );
          }
          newTable.appendChild( newColumn );
        }
      }
      //no CREATE TABLE statement
      else {
        Element newColumn;
        for( int i = 0; i < allColumnNames.length; i++ ) {
          newColumn = document.createElement( allColumnNames[i] );
          for( int j = 0; j < columnNames.length; j++ ) {
            if( allColumnNames[i].equalsIgnoreCase( columnNames[j] ) ){
              if( columnValues[j].equals("null") )
                newColumn.appendChild( document.createTextNode( "" ) );
              else
                newColumn.appendChild( document.createTextNode( columnValues[j] ) );
            }
            else {
              newColumn.appendChild( document.createTextNode("") );
            }
          }
          newTable.appendChild( newColumn );
        }
      }

      SearchElement newTableHash = new SearchElement( newTable );
      NodeList tables = searchDocument.getSubElementsByTagName("dml/"+tableName);
      Element before = null;
      if( tables.getLength() != 0 ) {
        before = (Element)tables.item(0);
        (new SearchElement(searchDocument.getSubElementsByTagName("dml").item(0))).insertBefore( newTableHash , before );
      } else {
        searchDocument.getSubElementsByTagName("dml").item(0).insertBefore( newTableHash , null );
      }
      saveDOM();
      } catch( Exception e ) {
        this.isError = true;
        throw new SQLException("Error in insert data : "+e.getMessage());
      }
  }

  /**
   * Update row in in XML file.
   *
   * @param tableName name of table which will be updatad.
   * @param columnNames names of columns in which will be added data.
   * @param columnValues value which will be insert into table.
   * @param whereColumnNames names of columns in WHERE clause.
   * @param whereColumnValues values of columns in WHERE clause.
   * @throws SQLException
   */
  protected void update(String tableName, String[] columnNames , String[] columnValues , String[] whereColumnNames , String[] whereColumnValues ) throws SQLException {
    try {
      boolean hasCreateTable =
          !this.getTableProperties( tableName ).get(1).toString().equalsIgnoreCase("NO CREATE TABLE");
      if( hasCreateTable ) {
        //check primary key column
        String[] primaryKeys = (String[])this.getTableProperties( tableName ).get(1);
        boolean isPrimarykey = false;
        for(int i = 0; i < primaryKeys.length; i++) {
          if( columnNames[0].equals( primaryKeys[i] ) )
            isPrimarykey = true;
        }
        if( isPrimarykey ) {
          NodeList columns = searchDocument.getSubElementsByCondition("dml/"+tableName+"/"+columnNames[0]+"="+columnValues[0]);
          if( columns.getLength() != 0 ) {
            this.isError = true;
            throw new SQLException("Can not insert duplicate value in primarykey column "+columnNames[0]+" !");
          }
        }
      }
      NodeList tableRows = this.searchDocument.getSubElementsByTagName("dml/"+tableName);
      //check if table row match conditions
      for(int i = 0; i < tableRows.getLength(); i++) {
        boolean isMatch = true;
        if( whereColumnNames != null && whereColumnValues != null ) {
          for(int k = 0; k < whereColumnNames.length; k++) {
            NodeList columns = ( (SearchElement)tableRows.item(i) ).getSubElementsByCondition(whereColumnNames[k]+"="+whereColumnValues[k]);
            if( columns.getLength() == 0 )
              isMatch = false;
          }
        }
        if( isMatch ) {
          for( int k = 0; k < columnNames.length; k++ ) {
            NodeList columns = ( (SearchElement)tableRows.item(i) ).getSubElementsByTagName(columnNames[k]);
            String[] notNullCols = null;
            if( hasCreateTable )
              notNullCols = (String[])this.getTableProperties( tableName ).get(2);
            if( columns.getLength() == 0 ) {
              //if column tag do not exist in row
              Element newColumn = document.createElement( columnNames[k] );
              if( columnValues[k].equals("null") ) {
                if( hasCreateTable ) {
                  for( int ii=0; ii < notNullCols.length; ii++ ) {
                    if( notNullCols[ii].equalsIgnoreCase(columnNames[k]) )
                      throw new SQLException("Column '" + columnNames[k] + "' can not be NULL.");
                  }
                }
                if ( hasCreateTable )
                  continue;      //do not add empty column
                else
                  newColumn.appendChild( document.createTextNode( "" ) );
              }
              else
                newColumn.appendChild( document.createTextNode( columnValues[k] ) );
              tableRows.item(i).appendChild( new SearchElement(newColumn) );
            }
            else {
              //if column tag exist
              SearchElement column = (SearchElement)columns.item(0);
              Node textNode = column.getFirstChild();
              if( textNode == null ) {
                Element newColumn = document.createElement( columnNames[k] );
                if( columnValues[k].equals("null") ) {
                  if( hasCreateTable ) {
                    for( int ii=0; ii < notNullCols.length; ii++ ) {
                      if( notNullCols[ii].equalsIgnoreCase(columnNames[k]) )
                        throw new SQLException("Column '" + columnNames[k] + "' can not be NULL.");
                    }
                  }
                  //when has CREATE TABLE statement, remove tag with null value
                  if( hasCreateTable )
                    column.getParentNode().removeChild(column);
                  else
                    newColumn.appendChild( document.createTextNode( "" ) );
                }
                //when new value is not null
                else {
                  newColumn.appendChild( document.createTextNode( columnValues[k] ) );
                  SearchElement parent = new SearchElement( column.getParentNode() );
                  parent.replaceChild( new SearchElement(newColumn) , column );
                }
              }
              else {
                if( columnValues[k].equals("null") ) {
                  if( hasCreateTable ) {
                    for( int ii=0; ii < notNullCols.length; ii++ ) {
                      if( notNullCols[ii].equalsIgnoreCase(columnNames[k]) )
                        throw new SQLException("Column '" + columnNames[k] + "' can not be NULL.");
                    }
                  }
                  //when has CREATE TABLE statement, remove tag with null value
                  if( hasCreateTable )
                    column.getParentNode().removeChild(column);
                  else
                    column.getFirstChild().setNodeValue( "" );
                }
                else
                  column.getFirstChild().setNodeValue( columnValues[k] );
              }
            }
          }
        }
      }
      saveDOM();
      } catch( Exception e ) {
        this.isError = true;
        throw new SQLException("Error in update data : "+e.getMessage());
      }
  }

  /**
   * Gets table properties in form ArrayList.
   * ArrayList[0] is string array with ALL column names in table.
   * ArrayList[1] is string array with colmn names which are PRIMARYKEYs.
   * ArrayList[2] is string array with colmn names which can not be NULL.
   *
   * If table has no CREATE TABLE statement , ArrayList[1] will have value "NO CREATE TABLE"
   *
   * @param tableName name of table.
   * @return list of table properties.
   * @throws SQLException
   */
  public ArrayList getTableProperties(String tableName) throws SQLException {
    ArrayList properties = new ArrayList();
    NodeList sqlStatements = searchDocument.getSubElementsByTagName("ddl");
      XmlSqlParser parser = new XmlSqlParser();
      for( int i = 0; i < sqlStatements.getLength(); i++ ) {
        Node node = sqlStatements.item(i);
        try {
          parser.parse( node.getFirstChild().toString() );
        } catch(Exception e) {
          this.isError = true;
          throw new SQLException("Error in parsing statement : "+e.getMessage());
        }
        if ( parser.getTableName().equalsIgnoreCase( tableName ) && parser.sqlType.equalsIgnoreCase( parser.CREATE_TABLE ) ) {
          properties.add( parser.getColumnNames() );
          properties.add( parser.getPrimaryKeys() );
          properties.add( parser.getNotnullColumns() );
        }
      }
      // no CREATE TABLE statement
      if( properties.size() == 0 ) {
        NodeList tableTags = searchDocument.getSubElementsByTagName("dml/"+tableName);
        if( tableTags.getLength() == 0 ) {
          this.isError = true;
          throw new SQLException("No existing table with specified name : "+tableName);
        }
        if( tableTags.getLength() != 0 ) {
          NodeList tableColumns = tableTags.item(0).getChildNodes();
          ArrayList arrColumns = new ArrayList();
          for( int i = 0; i < tableColumns.getLength(); i++ ) {
            String value = tableColumns.item(i).getNodeName();
            if( value != null && !value.equalsIgnoreCase("#text") ) {
              arrColumns.add( tableColumns.item(i).getNodeName() );
            }
          }
            properties.add( arrColumns.toArray(new String[0]) );
            properties.add( "NO CREATE TABLE" );
        }
      }
    return properties;
  }

  /**
   * Method is called when create new database file.
   *
   * @throws SQLException
   */
  protected void createDatabase()  throws SQLException {
      try {
        Properties prop = new Properties();
        prop.setProperty("version","1.0");
        prop.setProperty("encoding","ISO-8859-1");
        XMLDocumentFactory.serialize( this.searchDocument , fileName , prop );
      } catch( Exception e ) {
        throw new SQLException("Error in saving DOM  : "+e.getMessage());
      }
  }

  /**
   * Save DOM as XML file.
   *
   * @throws SQLException
   */
  protected void saveDOM()  throws SQLException {
    if( !XmlWriter.isError ) {
      try {
        Properties prop = new Properties();
        prop.setProperty("version","1.0");
        prop.setProperty("encoding","ISO-8859-1");
        if( this.autoCommit ) {
          XMLDocumentFactory.serialize( this.searchDocument , fileName , prop );
        }
        else
          this.searchDocumentStatic = searchDocument;

      } catch( Exception e ) {
        this.isError = true;
        throw new SQLException("Error in saving DOM  : "+e.getMessage());
      }
    }
  }

  /**
   * Method is used for saving DOM in xml file from connection object,when XmlConnection.commit() method
   * is called.
   * @param fileName full path of xml file.
   * @throws SQLException
   */
  public static void commit(String fileName)  throws SQLException {
    if( !XmlWriter.isError && XmlWriter.searchDocumentStatic != null ) {
      try {
        Properties prop = new Properties();
        prop.setProperty("version","1.0");
        prop.setProperty("encoding","ISO-8859-1");
        XMLDocumentFactory.serialize( XmlWriter.searchDocumentStatic , fileName , prop );
        XmlWriter.searchDocumentStatic = null;
      } catch( Exception e ) {
        throw new SQLException("Error in saving DOM  : "+e.getMessage());
      }
    }
  }

}

