/*
    Copyright (C) 2003  Together

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package org.enhydra.xml;

import java.io.File;
import java.io.FileOutputStream;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.HashMap;
import java.util.HashMap;

import org.w3c.dom.Attr;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.DOMException;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * @author Tweety
 *
 * A class representing a node in a meta-data tree, which implements
 * the <a href="../../../../api/org/w3c/dom/Element.html">
 *
 * <p> Namespaces are ignored in this implementation.  The terms "tag
 * name" and "node name" are always considered to be synonymous.
 *
 * @version 1.0
 */
public class ElementImpl extends NodeImpl implements Element {

	/**
	 * A <code>HashMap</code> of <code>AttrImpl</code> nodes representing
	 * attributes.
	 */
	protected HashMap attributes = null;


	/**
	 * Constructs an empty <code>ElementImpl</code>.
	 */
	public ElementImpl() {
		attributes = new HashMap();
		type = ELEMENT_NODE;
	}


	/**
	 * Constructs a <code>ElementImpl</code> from the given node,
	 * without creating entire children subtree.
	 *
	 * @param element , as a <code>ElementImpl</code>.
	 */
	public ElementImpl(ElementImpl element) {
		super(element);
		attributes = element.attributes;
		type = ELEMENT_NODE;
	}


	/**
	 * Constructs an <code>ElementImpl</code> with the given
	 * document owner and node name.
	 *
	 * @param ownerDoc the document owner of the node, as a <code>Document</code>.
	 * @param name is the name of the node, as a <code>String</code>.
	 */
	public ElementImpl(Document ownerDoc, String name) {
		super(ownerDoc,name,ELEMENT_NODE);
		this.attributes = new HashMap();
	}


	/**
	 * Constructs an <code>ElementImpl</code> with the given
	 * document owner, node name, node type and node value.
	 *
	 * @param ownerDoc the document owner of the node, as a <code>Document</code>.
	 * @param nodeName the name of the node, as a <code>String</code>.
	 * @param type the type of the node, as a <code>short</code>.
	 * @param value the value of the node, as a <code>String</code>.
	 */
	protected ElementImpl(Document ownerDoc, String nodeName, short type, String value) {
		super(ownerDoc, nodeName, type, value);
	}


	/**
	 * Constructs an <code>ElementImpl</code> from a given node (creates the children subtree too),
	 * as a <code>Node</code>
	 *
	 * @param node , as a <code>Node</code>.
	 */
	public ElementImpl(Node node) {
	    this(node,true);
	}


	/**
	 * Constructs an <code>ElementImpl</code> from a given node, as a <code>Node</code>,
	 * and deep as <code>boolean</code>.
	 *
	 * @param node , as a <code>Node</code>.
	 * @param deep if <code>true</code>, recursively clone the subtree
	 * under the specified node; if <code>false</code>, clone only the
	 * node itself.
	 */
	public ElementImpl(Node node, boolean deep) {
		super(node,false);
		attributes = new HashMap();
		NamedNodeMap attrs = node.getAttributes();
        if (attrs != null) {
    		for (int i = 0; i < attrs.getLength(); i++) {
    			Attr attr = new AttrImpl((Attr) attrs.item(i));
    			attributes.put(attr.getName(), attr);
    		}
        }
		if (deep)
		    initNodeImplChildren(node);
	}


	/**
	 * Creates new instance of <code>ElementImpl</code> from a given document
	 * as a <code>Document</code>.
	 *
	 * @param document document.
	 *
	 * @return new <code>Element</code> node as a root of the <code>Document</code>.
	 */
	public static Element newInstance(Document document) {
		Node root = document.getDocumentElement();
		return new ElementImpl(root);
	}


	/**
	 * Inserts the node <code>newChild</code> before the existing
	 * child node <code>refChild</code>. If <code>refChild</code> is
	 * <code>null</code>, insert <code>newChild</code> at the end of
	 * the list of children.
	 *
	 * @param newChild the <code>Node</code> to insert.
	 * @param refChild the reference <code>Node</code>.
	 *
	 * @return the node being inserted.
	 *
	 * @exception IllegalArgumentException if <code>newChild</code> is
	 * <code>null</code>.
	 */
	public Node insertBefore(Node newChild, Node refChild) {
        super.insertBefore(newChild,refChild);
		return newChild;
	}


	/**
	 * Replaces the child node <code>oldChild</code> with
	 * <code>newChild</code> in the list of children, and returns the
	 * <code>oldChild</code> node.
	 *
	 * @param newChild the <code>Node</code> to insert.
	 * @param oldChild the <code>Node</code> to be replaced.
	 *
	 * @return the node replaced.
	 *
	 * @exception IllegalArgumentException if <code>newChild</code> is
	 * <code>null</code>.
	 */
	public Node replaceChild(Node newChild, Node oldChild) {
		super.replaceChild(newChild,oldChild);
		return oldChild;
	}


	/**
	 * Removes the child node indicated by <code>oldChild</code> from
	 * the list of children, and returns it.
	 *
	 * @param oldChild the <code>Node</code> to be removed.
	 *
	 * @return the node removed.
	 *
	 * @exception IllegalArgumentException if <code>oldChild</code> is
	 * <code>null</code>.
	 */
	public Node removeChild(Node oldChild) {
        super.removeChild(oldChild);
		return oldChild;
	}


	/**
	 * Returns a duplicate of this node.  The duplicate node has no
	 * parent (<code>getParentNode</code> returns <code>null</code>).
	 * If a shallow clone is being performed (<code>deep</code> is
	 * <code>false</code>), the new node will not have any children or
	 * siblings.  If a deep clone is being performed, the new node
	 * will form the root of a complete cloned subtree.
	 *
	 * @param deep if <code>true</code>, recursively clone the subtree
	 * under the specified node; if <code>false</code>, clone only the
	 * node itself.
	 *
	 * @return the duplicate node.
	 */
	public Node cloneNode(boolean deep) {
		return new ElementImpl(this,deep);
	}





	// Methods from Element


	/**
	 * Returns tag name of this node.
	 *
	 * @return tag name of this node as a <code>String</code>.
	 */
	public String getTagName() {
		return nodeName;
	}


	/**
	 * Returns all attribute nodes of this node.
	 *
	 * @return all attribute nodes of this node as a <code>NamedNodeMap</code>.
	 */
	public NamedNodeMap getAttributes() {
		return new HashMapNamedNodeMap(attributes);
	}


	/**
	 * Returns the value of the attribute with given name.
	 *
	 * @param name name of attribute.
     *
	 * @return value of attribute.
	 */
	public String getAttribute(String name) {
		Attr attr = getAttributeNode(name);
		if (attr == null) {
			return "";
		}
		return attr.getValue();
	}


	/**
	 * Equivalent to <code>getAttribute(localName)</code>.
	 *
	 * @see #setAttributeNS
         * @param namespaceURI is name space
         * @param localName is string
         * @return node
	 */
	public String getAttributeNS(String namespaceURI, String localName) {
		return getAttribute(localName);
	}


	/**
	 * To the <code>name</code> attribute set value to <code>value</code>.
	 *
	 * @param name attribute value.
	 * @param value new attribute value.
	 */
	public void setAttribute(String name, String value) {
		// Note minor dependency on Crimson package
		// Steal the code if Crimson ever goes away

                // Code below does not work with jdk1.3
                /*
		if (!org.apache.crimson.util.XmlNames.isName(name)) {
			throw new NodeDOMException(
				DOMException.INVALID_CHARACTER_ERR,
				"Attribute name is illegal!");
		}
              */
		attributes.put(name, new AttrImpl(this, name, value));
	}


	/**
	 * Equivalent to <code>setAttribute(qualifiedName, value)</code>.
	 *
	 * @see #getAttributeNS
         * @param namespaceURI is name space of the node
         * @param qualifiedName is string
         * @param value is value of the node
	 */
	public void setAttributeNS(String namespaceURI, String qualifiedName, String value) {
		setAttribute(qualifiedName, value);
	}


	/**
	 * Removes attribute with the given name.
	 *
	 * @param name attribute name.
	 */
	public void removeAttribute(String name) {
		if (type != ELEMENT_NODE)
			throw new NodeDOMException(
				DOMException.NOT_SUPPORTED_ERR,
				"Node doesn't have attributes");
		removeAttribute(name, true);
	}


	private void removeAttribute(String name, boolean checkPresent) {
		if (attributes.remove(name) != null)
			return;
		// If we get here, the attribute doesn't exist
		if (checkPresent) {
			throw new NodeDOMException(
				DOMException.NOT_FOUND_ERR,
				"No such attribute!");
		}
	}


	/**
	 * Returns <code>true</code>, if this node has attributes, otherwise
	 * <code>false</code>.
	 *
	 * @return <code>true</code> if node has attributes, otherwise <code>false</code>..
	 */
	public boolean hasAttributes() {
		return attributes.size() > 0;
	}


	/**
	 * Returns <code>true</code>, if this node has attribute with given name,
	 * otherwise <code>false</code>.
	 *
	 * @return <code>true</code> if node has given attribute, otherwise <code>false</code>..
	 * @param name is name of the node
         */
	public boolean hasAttribute(String name) {
		return getAttributeNode(name) != null;
	}


	/**
	 * Equivalent to <code>removeAttribute(localName)</code>.
         * @param namespaceURI is name space of the node
         * @param localName is name of the node
	 */
	public void removeAttributeNS(String namespaceURI, String localName) {
		removeAttribute(localName);
	}


	/**
	 * Returns attribute value with given name of this node.
	 *
	 * @param name name of attribute.
     *
	 * @return value of attribute.
	 */
	public Attr getAttributeNode(String name) {
		return (Attr) attributes.get(name);
	}


	/**
	 * Equivalent to <code>getAttributeNode(localName)</code>.
	 *
	 * @see #setAttributeNodeNS
         * @param namespaceURI is name space of the node
         * @param localName is name of the node
         * @return node
	 */
	public Attr getAttributeNodeNS(String namespaceURI, String localName) {
		return getAttributeNode(localName);
	}


	/**
	 * Add new attribute to this node.
	 *
	 * @param newAttr new attribute.
     *
	 * @return new attribute as <code>AttrImpl</code>
         * @throws DOMException
	 */
	public Attr setAttributeNode(Attr newAttr) throws DOMException {
		AttrImpl attr;
		if (newAttr instanceof AttrImpl) {
			attr = (AttrImpl) newAttr;
		} else {
			attr = new AttrImpl(newAttr);
		}
		attributes.put(attr.getName(), attr);
		return attr;
	}


	/**
	 * Equivalent to <code>setAttributeNode(newAttr)</code>.
	 *
	 * @see #getAttributeNodeNS
         * @param newAttr is attribute of the node
         * @return node
	 */
	public Attr setAttributeNodeNS(Attr newAttr) {
		return setAttributeNode(newAttr);
	}


	/**
	 * Remove attribute from this node.
	 *
	 * @param oldAttr attribute that will be removed.
     *
	 * @return old attribute as <code>AttrImpl</code>.
	 */
	public Attr removeAttributeNode(Attr oldAttr) {
		removeAttribute(oldAttr.getName());
		return oldAttr;
	}


	/**
	 * Equivalent to <code>hasAttribute(localName)</code>.
         * @param namespaceURI is name space of the node
         * @param localName is name of the node
         * @return node
	 */
	public boolean hasAttributeNS(String namespaceURI, String localName) {
		return hasAttribute(localName);
	}


	/**
	 * Returns all <code>Element</code> nodes with given name,
	 * searching by all sub nodes from this node.
	 *
	 * @param name tag name.
     *
	 * @return all <code>Element</code> vith given name as <code>NodeList</code>.
	 */
    public NodeList getElementsByTagName(String name) {
        List list = new ArrayList();
        getElementsByTagName(name, list);
        return new NodeListImpl(list);
    }


    private void getElementsByTagName(String name, List list) {
        if (nodeName.equals(name)) {
            list.add(this);
        }

        Node child = getFirstChild();
        while (child != null) {
            if (child.getNodeType() == Node.ELEMENT_NODE)
                ((ElementImpl)child).getElementsByTagName(name, list);
            child = child.getNextSibling();
        }
    }


	/**
	 * Equivalent to <code>getElementsByTagName(localName)</code>.
         * @param namespaceURI is name space of the node
         * @param localName is name of the node
         * @return node
	 */
	public NodeList getElementsByTagNameNS(String namespaceURI,	String localName) {
		return getElementsByTagName(localName);
	}


	/**
	 * Returns <code>true</code> if this node has children nodes.
	 *
	 * @return <code>true</code> if this node has children.
	 */
	public boolean hasElementChildNodes() {
        Node child = getFirstChild();
        while (child != null) {
            if (child.getNodeType() == Node.ELEMENT_NODE)
                return true;
            child = child.getNextSibling();
        }
		return false;
	}


	/**
	 * Method beginToString for this class writes the xml
	 * begining tag string and all attributes.
	 *
	 * @param sb string buffer to add resulting string.
	 * @param indent used in formating the output.
	 */
	protected void beginToString(StringBuffer sb, Indent indent) {
		sb.append("\n" + indent + "<" + this.nodeName);

		for (Iterator iter = attributes.values().iterator(); iter.hasNext();) {
			Attr attr = (Attr) iter.next();
			sb.append(" " + attr.getNodeName() + "=\"" + attr.getNodeValue() + "\"");
		}
//        if (hasChildNodes()) {
       		sb.append(">");
        	indent.increment();
//        } else
//       		sb.append("/>");
	}

        /**
         * Check that the node is either <code>null</code> or an
         * <code>NodeImpl</code>.
         *
         * @exception DOMException if node is not an instance of <code>NodeImpl</code>.
         */
//        protected void checkNode(Node node) throws DOMException {
//                if (node == null) {
//                        return;
//                }
//                if (!(node instanceof ElementImpl))
//                        throw new NodeDOMException(DOMException.WRONG_DOCUMENT_ERR, "Node is not an instance of ElementImpl!");
//	}

	/**
	 * Method endToString for this class writes the xml
	 * ending tag string.
	 *
	 * @param sb string buffer to add resulting string.
	 * @param indent used in formating the output.
	 */
	protected void endToString(StringBuffer sb, Indent indent) {
//        if (hasChildNodes()) {
        	indent.decrement();
        	if (hasElementChildNodes())
			    sb.append("\n" + indent + "</" + this.nodeName + ">");
			else
			    sb.append("</" + this.nodeName + ">");
//        }
	}
}
